from .ip_to_int import ip_to_int, int_to_ip


class IP:
	def __init__(self, x):
		if isinstance(x, str):
			if '.' not in x:
				x = int(x)

		if isinstance(x, IP):
			self._int = x._int
			self._string = x._string

		elif isinstance(x, str):
			self._string = x
			self._int = ip_to_int(x)

		elif isinstance(x, int):
			self._int = x
			self._string = int_to_ip(x)

		else:
			raise ValueError(f'Cannot accept {x} of type {type(x)}')

	@property
	def int(self):
		"""
		:rtype: int
		"""
		return self._int

	def __repr__(self):
		return f'ip:{self._string}'

	def __str__(self):
		return self._string

	def __lt__(self, other):
		if not isinstance(other, IP):
			other = IP(other)

		return self._int < other._int

	def __gt__(self, other):
		if not isinstance(other, IP):
			other = IP(other)
		return self._int > other._int

	def __eq__(self, other):
		if not isinstance(other, IP):
			other = IP(other)
		return self._int == other._int

	def __ge__(self, other):
		if not isinstance(other, IP):
			other = IP(other)
		return self._int >= other._int

	def __le__(self, other):
		if not isinstance(other, IP):
			other = IP(other)
		return self._int <= other._int

	def __ne__(self, other):
		if not isinstance(other, IP):
			other = IP(other)
		return self._int != other._int



from .ip_to_int import get_ip_range, ip_to_int


def is_in_ip_range(ip, start_ip, end_ip):
	ip_int = ip_to_int(ip)
	start_int = ip_to_int(start_ip)
	end_int = ip_to_int(end_ip)

	return start_int <= ip_int and ip_int <= end_int


def is_in_block(ip, cidr_block):
	start_ip, end_ip = get_ip_range(cidr_block)
	return is_in_ip_range(ip=ip, start_ip=start_ip, end_ip=end_ip)



def is_ipv4(x):
	if not isinstance(x, str):
		raise TypeError('is_ipv4 only works on strings!')
	if '.' not in x:
		return False

	if x.count('/') > 1:
		return False

	if x.count('.') != 3:
		return False

	if x.count('/') == 1:
		x, bits = x.split('/')
		try:
			if int(bits) > 32 or int(bits) < 1:
				return False
		except ValueError:
			return False

	try:
		elements = [int(e) for e in x.split('.')]
		return all([e >= 0 and e <= 255 for e in elements])
	except ValueError:
		return False



def reverse_insort(a, x, lo=0, hi=None):
    """Insert item x in list a, and keep it reverse-sorted assuming a
    is reverse-sorted.

    If x is already in a, insert it to the right of the rightmost x.

    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.
    """
    if lo < 0:
        raise ValueError('lo must be non-negative')
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        if x > a[mid]: hi = mid
        else: lo = mid+1
    a.insert(lo, x)



from setuptools import setup, find_packages


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='internet',
	version='2021.8.25',
	description='Python library for working with IP addresses and other internet related functionalities',
	long_description=readme(),
	long_description_content_type='text/markdown',
	url='https://github.com/idin/internet',
	author='Idin',
	author_email='py@idin.ca',
	license='MIT',
	packages=find_packages(exclude=("jupyter_tests", ".idea", ".git")),
	install_requires=['base32hex', 'numpy', 'pandas', 'pyspark', 'disk'],
	python_requires='~=3.6',
	zip_safe=False
)

# todo add ipv6 to the library



from .ip_to_int import get_ip_range
from .IP import IP
import bisect

class CidrBlock:
	def __init__(self, s):
		if isinstance(s, CidrBlock):
			self._start_ip = s.start
			self._end_ip = s.end
			self._bits = s.bits
		else:
			start_ip, end_ip = get_ip_range(s)
			self._bits = int(s.split('/')[1])
			self._start_ip = IP(start_ip)
			self._end_ip = IP(end_ip)

	@property
	def start(self):
		"""
		:rtype: IP
		"""
		return self._start_ip

	@property
	def end(self):
		"""
		:rtype: IP
		"""
		return self._end_ip

	@property
	def bits(self):
		"""
		:rtype: int
		"""
		return self._bits

	def contains(self, item):
		if isinstance(item, str):
			if '/' in item:
				item = CidrBlock(item)

		if isinstance(item, (CidrBlock, IPSpace)):
			return self.start <= item.start and item.end <= self.end

		if not isinstance(item, IP):
			item = IP(item)
		return self.start <= item and item <= self.end

	def is_adjacent_to(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		if other.start.int - self.end.int == 1:
			return True
		elif self.start.int - other.end.int == 1:
			return True
		else:
			return False

	@property
	def size(self):
		return 2 ** (32 - self.bits)

	def __str__(self):
		return f'{self.start}/{self.bits}'

	def __repr__(self):
		return f'CIDR Block: {str(self)}'

	def __lt__(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		return (self.bits, self.start) < (other.bits, other.start)

	def __contains__(self, item):
		return self.contains(item)

	def __gt__(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		return (self.bits, self.start) > (other.bits, other.start)

	def __le__(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		return (self.bits, self.start) <= (other.bits, other.start)

	def __ge__(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		return (self.bits, self.start) >= (other.bits, other.start)

	def __eq__(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		return (self.bits, self.start) == (other.bits, other.start)

	def __ne__(self, other):
		if not isinstance(other, CidrBlock):
			other = CidrBlock(other)
		return (self.bits, self.start) != (other.bits, other.start)

	def __add__(self, other):
		if not isinstance(other, (CidrBlock, IPSpace)):
			other = CidrBlock(other)

		if isinstance(other, CidrBlock):
			if self in other:
				return other
			elif other in self:
				return self
			else:
				return IPSpace(cidr_blocks=[self, other])
		else:
			return IPSpace(cidr_blocks=[self] + other.cidr_blocks)


class IPSpace:
	def __init__(self, cidr_blocks):
		"""
		:type cidr_blocks: list[CidrBlocks] or list[str]
		"""

		self._blocks = []
		for block in cidr_blocks:
			self.append(block)

			if not isinstance(block, CidrBlock):
				x = CidrBlock(block)
			else:
				x = block
			if x not in self._blocks:
				self._blocks.append(x)

	def __len__(self):
		return len(self._blocks)

	def __str__(self):
		return '\n'.join([str(block) for block in self.cidr_blocks])

	def __repr__(self):
		return f'IPSpace:\n{str(self)}'

	@property
	def start(self):
		"""
		:rtype: IP
		"""
		return min([block.start for block in self.cidr_blocks])

	@property
	def end(self):
		"""
		:rtype: IP
		"""
		return max([block.end for block in self.cidr_blocks])

	def append(self, cidr_block):
		"""
		:type cidr_block: CidrBlock or str
		"""
		if not isinstance(cidr_block, CidrBlock):
			cidr_block = CidrBlock(cidr_block)

		if self.contains(cidr_block):
			pass

		else:
			# if cidr_block contains any block, it should replace it
			blocks = []
			for i in range(len(self)):
				item = self.cidr_blocks[i]
				if not cidr_block.contains(item):
					bisect.insort(blocks, item)
			bisect.insort(blocks, cidr_block)
			self._blocks = blocks

	@property
	def cidr_blocks(self):
		"""
		:rtype: list[CidrBlock]
		"""
		return self._blocks

	def contains(self, item):
		if isinstance(item, IPSpace):
			for item_block in item.cidr_blocks:
				if not self.__contains__(item_block):
					return False
			return True

		else:
			for block in self.cidr_blocks:
				if block.contains(item):
					return True
			return False

	def copy(self):
		result = IPSpace(cidr_blocks=[])
		for block in self.cidr_blocks:
			result._blocks.append(block)
		return result

	def __contains__(self, item):
		return self.contains(item)

	def __add__(self, other):
		result = self.copy()
		if isinstance(other, list):
			other = IPSpace(cidr_blocks=other)

		if isinstance(other, IPSpace):
			for block in other.cidr_blocks:
				result.append(block)
		else:
			other = CidrBlock(other)
			result.append(other)

		return result



def ip_to_int(ip):
	parts = ip.split('.')
	return (int(parts[0]) << 24) + (int(parts[1]) << 16) + (int(parts[2]) << 8) + int(parts[3])


def int_to_ip(x):
	return '.'.join([str(x >> (i << 3) & 0xFF) for i in range(4)[::-1]])


def get_cidr_block(ip_address, bits):
	"""
	reduces ip address to a cidr block
	:type ip_address: str or Column
	:type bits: int
	:rtype: str
	"""

	if bits >= 32:
		return ip_address
	reduced_bytes = (32 - bits) // 8
	reduced_bits = (32 - bits) % 8
	parts = ip_address.split('.')[:4 - reduced_bytes]

	parts[-1] = str((int(parts[-1]) >> reduced_bits) << reduced_bits)
	return f'{".".join(parts)}/{bits}'


def get_ip_range(cidr_block):
	"""
	gets a cidr block of with cidr format '24.0.0.0/12'
	returns a start and end ip address of format '24.0.0.0', '24.15.255.255'
	:param cidr_block:
	:return:
	"""
	start, bits = cidr_block.split('/')
	start = start.strip()
	bits = int(bits.strip())

	start_int = ip_to_int(start)
	end_int = start_int + 2 ** (32 - bits) - 1
	end = int_to_ip(end_int)
	return start, end
